
#include "usb910conf.h"

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <avr/sleep.h>

#include "usbcdc.h"
#include "usb910.h"

#define MAX_55mA

#ifndef ADCD
#define ADCD ADC
#endif

#define poll    usbcdc_poll

#define bit_set(port,bit)	{ port |= (1<<(bit));}
#define bit_clear(port,bit)	{ port &= ~(1<<(bit));}
#define min(a,b)		((a)<(b) ? (a):(b))


#ifdef ADC_EXTENSION


#define ADC_MAX_CONF	1
static uint8_t adc_conf[ADC_MAX_CONF];
/* result format 
 *
 *  int16_t: unit 0.15625 mV (10/64 mV)
 *
 *   +-+-+----------+---+
 *   |1|1|  10      | 4 |
 *   +-+-+----------+---+
 *       |          |
 *
 */

#if defined(__AVR_ATtiny44__) || defined(__AVR_ATtiny84__)

/* 
 *   VCC     --  ADC3 x k (about 5/2)
 *   CURRENT --  DELTA ADC0(+) ADC1(-)
 */

#define REF_VCC		(0<<REFS0)
#define REF_V110	(2<<REFS0)
#define CHAN_ADC2	0x02
#define CHAN_ADC3	0x03
#define CHAN_TEMP	0x22

#define GAIN_x1		0
#define GAIN_x20	1

#define DELTA_ADC32	0x30 
#define DELTA_ADC23	0x20 
#define DELTA_ADC33	0x24 
#define CHAN_CURRENT	DELTA_ADC32

#define SEL_RANGE_V005  (uint16_t)( 0x4000 * 1.1/2.56 /20.0 *  0.93)

#else
#error "ADC_EXTENSION is not supported\n"
#endif

static inline uint8_t adc_is_busy() {
	return bit_is_set(ADCSRA,ADSC);
}

#define ALWAYS_ADEN

static uint16_t do_adc_native(uint8_t admux) {
	uint16_t r1,r2,r3,tmp;
	ADMUX = admux;
#ifdef ALWAYS_ADEN
	ADCSRA |= (1<<ADSC);
	while (adc_is_busy()) {
		poll();
	}
	r1 = ADCD;
	ADCSRA |= (1<<ADSC);
	while (adc_is_busy()) {
		poll();
	}
	r2 = ADCD;
	ADCSRA |= (1<<ADSC);
	while (adc_is_busy()) {
		poll();
	}
	r3 = ADCD;
	if (r2 > r1) {
		tmp = r1; r1 = r2; r2 =tmp;
	}
	if (r3 > r1) {
		tmp = r1; r1 = r3; r3 =tmp;
	}
	if (r3 > r2) {
		tmp = r2; r2 = r3; r3 =tmp;
	}
	return r2;
#else
	ADCSRA |= (1<<ADEN);
	ADCSRA |= (1<<ADSC);
	while (adc_is_busy()) {
		poll();
	}
	/* discard 1st result */
	ADCSRA |= (1<<ADSC);
	while (adc_is_busy()) {
		poll();
	}
	ADCSRA &= ~(1<<ADEN);
	return ADCD;
#endif
}

void init_adc() {
	/* setup ADC @ 1/64 clock */
#ifdef ALWAYS_ADEN
	ADCSRA = (6<<ADPS0) | (1<<ADEN);
#else
	ADCSRA = (6<<ADPS0);
#endif
	/* digial input disable */
	bit_clear(DDRA, PA2);
	bit_clear(DDRA, PA3);
	bit_clear(PORTA, PA2);
	bit_clear(PORTA, PA3);
	DIDR0 = (1<<ADC2D) | (1<<ADC3D) ;
}

#if 0
void term_adc() {
	ADCSRA = 0;
	ADCSRB = 0;
	DIDR0 = 0;
}
#endif

int8_t adc_send(uint8_t size) {
#if ADC_MAX_CONF == 1
	if (size < 1) return 0;
	adc_conf[0] = ext_buf[0];
	return 1;
#else
	uint8_t i;
	if (size > ADC_MAX_CONF) size = ADC_MAX_CONF;
	for (i=0; i<size; i++) {
		adc_conf[i] = ext_buf[i];
	}
	return i;
#endif
}


static inline uint16_t temp_val() {
	return do_adc_native((2<<REFS0) | CHAN_TEMP);
}

static inline uint16_t current_val() {
	uint16_t result;
#ifdef MAX_55mA
	result = do_adc_native((2<<REFS0) | ( CHAN_CURRENT + GAIN_x20));
	/* same as x 11 / 32 */
	result = (result << 3) + (result << 1) + result;
	result >>= 5;
#else
	result = do_adc_native((2<<REFS0) | ( CHAN_CURRENT + GAIN_x1));
#if 0
	result *= 55;
	result >>= 3;
#else
	result = (result<<6) - (result<<3) - result;
	result >>= 3;
#endif
#endif
	return result;
}

int8_t adc_recv(uint8_t size) {
#if ADC_MAX_CONF == 1
	if (size < 2) return 0;
	if (adc_conf[0] & ADC_CHAN_DELTA) {
		    ((uint16_t *)ext_buf)[0] = current_val();
	} else {
		    ((uint16_t *)ext_buf)[0] = temp_val();
	}
	return 2;
#else
	uint8_t i;
	size >>= 1;

	for (i=0; i<size; i++) {
	    if (adc_conf[i] & ADC_CHAN_DELTA) {
		    ((uint16_t *)ext_buf)[i] = current_val();
	    } else {
		    ((uint16_t *)ext_buf)[i] = temp_val();
            }
	}
	return (i<<1);
#endif
}

#endif
