/*
 * Copyright (c) 2007 Koji Suzuki
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <time.h>
#include <ctype.h>
#include <unistd.h>


#if defined(_WIN32)
#define WIN32NATIVE
#endif
#if defined(WIN32NATIVE)
#include <windows.h>
#include <io.h>
//#include <mmsystem.h>

static inline double dtime() {
	return (double)timeGetTime()/1.0e3;
}

void usleep(int t) {
	Sleep(1 + (t+999)/1000);
}


int set_serial(int fd) {
	HANDLE handle;
	DCB dcb;
	COMMTIMEOUTS cto;

	handle = (HANDLE)(void *)_get_osfhandle(fd);
	if (handle == INVALID_HANDLE_VALUE) {
		printf("get_osfhandle error\n");
		fflush(stdout);
		exit(1);
		//return -1;
	}
	if (!SetupComm(handle, 256, 256)) {
		return -1;
	}
	memset(&dcb, 0,  sizeof(DCB));
	dcb.DCBlength = sizeof(DCB);
	dcb.BaudRate = CBR_19200;
	dcb.fBinary = 1;
	dcb.fDtrControl = DTR_CONTROL_DISABLE;
	dcb.fRtsControl = RTS_CONTROL_DISABLE;
	dcb.ByteSize = 8;
	dcb.Parity = NOPARITY;
	dcb.StopBits = ONESTOPBIT;
	if (!SetCommState(handle, &dcb)) {
		return -1;
	}
	memset(&cto, 0, sizeof(COMMTIMEOUTS));
	cto.ReadIntervalTimeout = 500;        // 200 ms
	cto.ReadTotalTimeoutMultiplier = 500; // 200 ms
	cto.ReadTotalTimeoutConstant = 500;   // 200 ms
	SetCommTimeouts(handle, &cto);
	return 0;
}

#else
#include <termios.h>
#include <sys/time.h>
#include <sys/types.h>

static inline double dtime() {
	struct timeval tv;
	double t;

	gettimeofday(&tv,0);
	return (double)(unsigned long)tv.tv_sec + (double)tv.tv_usec/1.0e6;
}

int set_serial(int fd) {
	int speed;
	struct termios term;

	tcgetattr(fd, &term);
	speed = B19200;
	cfsetispeed(&term,speed);
	cfsetospeed(&term,speed);
	cfmakeraw(&term);
	usleep(100000); /* 100 ms */
	tcsetattr(fd, TCSANOW, &term);
	/* setup done */
	return 0;
}

#ifndef O_BINARY
#define O_BINARY	0
#endif

#endif

#define DEV_CHAN	0x10
#define uint8_t unsigned char
#define int16_t short
#define uint16_t unsigned short

#define ADC_CHAN_AIN1           0x0 // using ADC4
#define ADC_CHAN_AIN2           0x1 // using ADC5
#define ADC_CHAN_AIN3           0x2 // using ADC6
#define ADC_CHAN_AIN4           0x3 // using ADC9
#define ADC_CHAN_AIN5           0x4 // internal: using ADC0
#define ADC_CHAN_AIN6           0x5 // internal: using ADC1
#define ADC_CHAN_VCC            0x6 // internal: using ADC3
#define ADC_CHAN_TEMP           0x7

#define ADC_CHAN_CURRENT        0x8 //internal
#define ADC_CHAN_AIN12          0x9
#define ADC_CHAN_AIN23          0xa
#define ADC_CHAN_AIN31          0xb
#define ADC_CHAN_AIN65          0xc //internal reverse input
#define ADC_CHAN_AIN21          0xd // reverse input
#define ADC_CHAN_AIN32          0xe // reverse input
#define ADC_CHAN_AIN13          0xf // reverse input

#define ADC_CHAN_DELTA          0x08
#define ADC_CHAN_MASK           0x0f

#define ADC_RANGE_AUTO          0x00 // Auto Range
#define ADC_RANGE_V512          0x10 // Reference VCC (+ calibrate VCC)
#define ADC_RANGE_V256          0x20 // Reference 2.56V
#define ADC_RANGE_V110          0x30 // Reference 1.10V
#define ADC_RANGE_V032          0x50 // Reference 2.56V + Gain x8
#define ADC_RANGE_V008          0x70 // Reference 2.56V + Gain x32
#define ADC_RANGE_MASK          0x70
#define ADC_BIPOLAR             0x80 // use BIN

#define ADC_MAX_CONF            4

void send_and_recv(int fd, uint8_t *buf, int slen, int rlen) {
	int r,i;
	if (slen != 0) {
	    r = write(fd, buf, slen);
	    if (r != slen ) {
		fprintf(stderr,"cmd send error r = %d %02x\n", r , buf[0]);
		exit(1);
	    }
	}
	i = 0;
	if (rlen != 0) {
re_read:
	    r = read(fd, buf+i , rlen-i);
	    if (r <= 0) {
		fprintf(stderr,"reply error r = %d %02x\n", r , buf[0]);
		exit(1);
	    }
	    if (r != rlen-i) {
		i += r;
		goto re_read;
	    }
	}
}

init_usb910(int fd) {
	int r,i;
	uint8_t buf[1024];

	for (i=0; i<10; i++) {
		r = write(fd, "S", 1);
		if (r < 1) {
			perror("write 1");
			exit(1);
		}
		usleep(1000); /* 1 ms */
		r = read(fd, buf, 7);
		if (r == 7 && !strncmp(buf,"USB",3)) {
			buf[7] = 0;
			printf("nego OK [%s]\n", buf);
			r = write(fd, "a", 1);
			usleep(1000); /* 1 ms */
			r = read(fd, buf, 1);
			if (r == 1 && buf[0] == 'Y') {
				printf("auto address incliment OK\n");
				goto next;
			} else {
				printf("auto address incliment NG\n");
			}
		}
		printf("read len = %d\n",r);
	}
	return -1;
next:
	buf[0] = 'e';
	send_and_recv(fd, buf, 1, 1);
	if (buf[0] != 'Y') {
		fprintf(stderr,"enter extended mode Error\n");
	}
}

void i2c_dev_write(int fd, int bytes, uint8_t *in) {
	uint8_t buf[1024];
	int8_t r;
	int i,try;

	for (try=0 ; try< 3; try++) {
		buf[0] = 's';
		buf[1] = DEV_CHAN;
		buf[2] = bytes;
		for (i=0; i<bytes; i++) {
			buf[3+i] = in[i];
		}
		send_and_recv(fd, buf, 3 + bytes,1);
		if ((int8_t)(buf[0]) != -2) {
			break;
		}
	}
	if ((int8_t)(buf[0]) < 0) {
		fprintf(stderr, "i2c_write error (1) %d \n",(int8_t)(buf[0]));
		exit(1);
	}
#if 0
	printf("i2c_dev_write bytes = %d result = %d\n",
				bytes, (int8_t)(buf[0]));
	fflush(stdout);
#endif
}

int i2c_dev_read(int fd, int bytes, uint8_t *out) {
	uint8_t buf[1024];
	int8_t r;
	int i,try,rbytes;

	for (try=0 ; try< 3; try++) {
		buf[0] = 'r';
		buf[1] = DEV_CHAN;
		buf[2] = bytes;
		send_and_recv(fd, buf, 3 , 1);
		if ((int8_t)(buf[0]) != -2) {
			break;
		}
	}
	if ((int8_t)(buf[0]) < 0) {
		fprintf(stderr, "i2c_read error (1) %d \n",(int8_t)(buf[0]));
		exit(1);
	}
	rbytes = (int8_t)(buf[0]);
#if 0
	printf("i2c_dev_read result bytes = %d \n", rbytes);
	fflush(stdout);
#endif
	for (i=0; i<bytes; i++) {
		out[i] = 0xff;
	}
	i = 0;
re_read:
	r = read(fd, out+i, rbytes-i);
	if (r < 0) {
		fprintf(stderr,"i2c_read error (2) %d\n",r);
		exit(1);
	}
	if (r < rbytes-i) {
		i += r;
		goto re_read;
	}
	return rbytes;
}

#define FCNT_START      0x01

int mult = 4;

start_fcnt(int fd) {
	int r,i;
        uint8_t buf[1024];
	int fnx,count,t0,t1;

	count = 1;
	buf[0] = 's';	// send
	buf[1] = 3;	// fcnt
	buf[2] = count;	// conf len
	buf[3] = FCNT_START;

	r = write(fd, buf, 3 + count);
	if (r != 3 + count) {
		fprintf(stderr,"write cmd error 1 r = %d \n", r);
		perror("write cmd 2");
		exit(1);
	}
	usleep(200000); /* 0.2 sec */
	r = read(fd, buf, 100);
	if (r != 1 && buf[0] != count) {
		fprintf(stderr,"reply error 1 r = %d r[0] = %02x\n", r, buf[0]);
		if (r < 0) perror("send reply");
		exit(1);
	}
	count = 8;
	usleep(10000); /* 10 ms */
	buf[0] = 'r';	// recv
	buf[1] = 3;	// fcnt
	buf[2] = count;	// result len
	r = write(fd, buf, 3);
	if (r != 3) {
		fprintf(stderr,"write cmd error 2 r = %d \n", r);
		perror("write cmd 3");
		exit(1);
	}
	usleep(1000); /* 1 ms */
	r = read(fd, buf, 100);
	if ((r < 3 + 1) && (buf[0] < 3)) {
		fprintf(stderr,"reply error 2 r = %d r[0] = %02x\n", r, buf[0]);
		exit(1);
	}
	fnx = buf[1];
	fnx |= (int)(buf[2]) << 8;
	fnx |= (int)(buf[3]) << 16;
	printf("fcnt result %10d x%d %10d\n",fnx,mult,fnx*mult);
}

char *conf_str(int n) {
     int chan;
     int range;
     static char buf[256];
     buf[0] = 0;

     chan = n & ADC_CHAN_MASK;
     switch (chan) {
	case ADC_CHAN_TEMP: strcpy(buf, "TEMP "); break;
	case ADC_CHAN_VCC : strcpy(buf, "VCC  "); break;
	case ADC_CHAN_AIN1: strcpy(buf, "AIN1 "); break;
	case ADC_CHAN_AIN2: strcpy(buf, "AIN2 "); break;
	case ADC_CHAN_AIN3: strcpy(buf, "AIN3 "); break;
	case ADC_CHAN_AIN4: strcpy(buf, "AIN4 "); break;
	case ADC_CHAN_AIN5: strcpy(buf, "AIN5 "); break;
	case ADC_CHAN_AIN6: strcpy(buf, "AIN6 "); break;
	case ADC_CHAN_CURRENT: strcpy(buf, "CURR "); break;

	case ADC_CHAN_AIN12: strcpy(buf, "AIN12"); break;
	case ADC_CHAN_AIN23: strcpy(buf, "AIN23"); break;
	case ADC_CHAN_AIN31: strcpy(buf, "AIN31"); break;
	case ADC_CHAN_AIN21: strcpy(buf, "AIN21"); break;
	case ADC_CHAN_AIN32: strcpy(buf, "AIN32"); break;
	case ADC_CHAN_AIN13: strcpy(buf, "AIN13"); break;
	default:
	case ADC_CHAN_AIN65: strcpy(buf, "AIN65"); break;
     }

     if (n & ADC_BIPOLAR)
     	strcat(buf," B ");
     else
     	strcat(buf,"   ");
     range = n & ADC_RANGE_MASK;
     switch (range) {
	case ADC_RANGE_AUTO: strcat(buf, "AUTO "); break;
	case ADC_RANGE_V512: strcat(buf, "V512 "); break;
	case ADC_RANGE_V256: strcat(buf, "V256 "); break;
	case ADC_RANGE_V110: strcat(buf, "V110 "); break;
	case ADC_RANGE_V032: strcat(buf, "V032 "); break;
	case ADC_RANGE_V008: strcat(buf, "V008 "); break;
     }
     return buf;
}

get_adc(int fd2, int count, uint8_t *in) {
	int r,i,rb,x;
        uint8_t buf[1024];
        uint8_t obuf[1024];
	int16_t *result;

	buf[0] = 's';	// send
	buf[1] = 1;	// adc
	buf[2] = count;	// conf len
	for (i=0; i< count; i++) {
		buf[3+i] = in[i];
	}
	send_and_recv(fd2, buf, 3 + count, 1);

   for (x=0; x<2; x++) {
	obuf[0] = 'r';	// recv
	obuf[1] = 1;	// adc
	obuf[2] = count*2;	// result len
	send_and_recv(fd2, obuf, 3 , 1);
	rb = obuf[0];
	if (rb != count*2) {
		fprintf(stderr,"adc read error count = %d result = %d\n",count,rb);
		exit(1);
	}
	i = 0;
	while (i<rb) {
		r = read(fd2, obuf+1+i, rb - i);
		if (r < 0) {
			fprintf(stderr,"adc read data error result = %d\n",r);
		}
		i += r;
	}
   }

	result = (uint16_t *)(obuf+1);
	for (i=0; i<count; i++) {
	    if ((in[i] & ADC_CHAN_MASK) == ADC_CHAN_TEMP) {
		printf("    %02x:%s: %07d 0x%04x %10.3f\n",
			in[i], conf_str(in[i]), result[i], result[i] & 0xffff,
			28 + 1.0 * (result[i] - 308));
	    } else {
		printf("    %02x:%s: %07d 0x%04x %10.3f\n",
			in[i], conf_str(in[i]), result[i], result[i] & 0xffff,
			10.0 * ((in[i]&ADC_BIPOLAR)?
				result[i]:(uint16_t)(result[i]))/64.0 );
	    }
	}
}

#if defined(WIN32NATIVE)
char *port_name = "COM4";
#else
char *port_name = "/dev/ttyACM0";
#endif

main(int argc,char *argv[]) {
	int fd,c,i,pwm,adc1;
	int log_cnt = 0;
	int cmd = -1;
	int cmd_arg = -1;
	uint8_t buf[1024];

	pwm = -1;
	while ((c=getopt(argc,argv,"P:p:l:o:m:xe")) != EOF) {
	    switch(c) {
		case 'P':
			port_name = optarg;
			break;
		case 'p':
			pwm = atoi(optarg);
			break;
		case 'l':
			log_cnt = atoi(optarg);
			break;
		case 'o':
		case 'm':
			cmd = c;
			cmd_arg = atoi(optarg);
			break;
		case 'x':
		case 'e':
			cmd = c;
			break;
	    }
	}
	fd = open(port_name, O_RDWR | O_BINARY);
	if (fd < 0) {
		printf("can't open serial\n");
		exit(1);
	}
	set_serial(fd);
	init_usb910(fd);

	if (cmd != -1) pwm = -1;
	if (cmd == 'x' || cmd == 'e') {
		buf[0] = cmd;
		i2c_dev_write(fd, 1, buf);
	}  else if (cmd == 'l' || cmd == 'm') {
		buf[0] = cmd;
		buf[1] = cmd_arg;
		i2c_dev_write(fd, 2, buf);
	}
	if (pwm > 0) {
		buf[0] = 'D';
		buf[1] = pwm;
		i2c_dev_write(fd, 2, buf);
	} else {
		buf[0] = 'd';
		i2c_dev_write(fd, 1, buf);
	}
	i2c_dev_read(fd, 8, buf);
	pwm = buf[0];

	buf[0] = 'S';
	i2c_dev_write(fd, 1, buf);
	usleep(100000); // 100 ms
	i2c_dev_read(fd, 8,buf);
	buf[7] = 0;
	printf("#Soft %s\n", buf);
	
	buf[0] = 'v';
	buf[1] = 2;
	i2c_dev_write(fd, 2, buf);
	i2c_dev_read(fd, 8, buf);
	adc1 = buf[0] + buf[1]*256;
	printf("pwm %2d  adc1 %5.3f \n", pwm , 2.488 * 3 /1024.0 * adc1);


	buf[0] = ADC_CHAN_VCC ;
	get_adc(fd, 1, buf);
	buf[0] = ADC_CHAN_TEMP ;
	get_adc(fd, 1, buf);
	buf[0] = ADC_CHAN_CURRENT |  ADC_RANGE_V008;
	get_adc(fd, 1, buf);
	for (i=0; i<log_cnt; i++) {
		buf[0] = 'v';
		buf[1] = 2;
		i2c_dev_write(fd, 2, buf);
		i2c_dev_read(fd, 8, buf);
		adc1 = buf[0] + buf[1]*256;
		printf("pwm %2d  adc1 %5.3f \n", pwm , 2.488 * 3/1024.0 * adc1);
	}
}
