/*
 * Copyright (c) 2007 Koji Suzuki
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>

#include "usiconf.h"
#include "i2cdev.h"

#define _USIBR	USIDR

#define bit_set(port,bit)	{ port |= (1<<(bit));}
#define bit_clear(port,bit)	{ port &= ~(1<<(bit));}

/* note : TX/RX mean TX/RX for HOST, RX/TX for DEVICE */

#ifndef I2CDEV_TX_BUFSIZE
#define I2CDEV_TX_BUFSIZE 8
#endif
#ifndef I2CDEV_RX_BUFSIZE
#define I2CDEV_RX_BUFSIZE 8
#endif

static uint8_t i2cdev_myaddr;

static uint8_t i2cdev_tx_buf[I2CDEV_TX_BUFSIZE];
volatile static uint8_t i2cdev_tx_putlen1; // recving data length
volatile static uint8_t i2cdev_tx_putlen2; // getc-able length
volatile static uint8_t i2cdev_tx_getlen; // used by getc

static uint8_t i2cdev_rx_buf[I2CDEV_RX_BUFSIZE];
volatile static uint8_t i2cdev_rx_putlen1; // used by putc
volatile static uint8_t i2cdev_rx_putlen2; // send-able length
volatile static uint8_t i2cdev_rx_getlen; // sending length

#define STATE_IDLE		0
#define STATE_RECV_ADDR		1
#define STATE_SEND_REPLY	2
#define STATE_RECV_DATA		3
#define STATE_SEND_DATA_START	4
#define STATE_RECV_REPLY	5
#define STATE_SEND_DATA		6

/* 
 * IDLE -> RECV_ADDR 
 *                  (WRITE COMMAND)
 *                   -> SEND_REPLY -> RECV_DATA -+
 *                         |                     |
 *                         +---<-----------------+
 *
 *                  (READ COMMAND)
 *                   -> SEND_DATA_START -> SEND_DATA -> RECV_REPLAY -+
 *                                             |                     |
 *                                             +---<-----------------+
 *
 * note: goto IDLE with any error /or complete
 *
 */

static uint8_t state;

/* 
 *          getc    return getc  1st putc   done getc     flush
 *           (A)        (B)        (C)         (D)          (E)
 * WRITE  ---------------|----------------------|---------------------
 *          Accept       |        Deny          |     Accept
 *                
 * READ ---------------------------|-------------------------|------
 *          old data               |       0xff              | new data
 */

void i2cdev_flush() {
	cli();
	i2cdev_tx_getlen = i2cdev_tx_putlen2; // flush input
	i2cdev_rx_putlen2 = i2cdev_rx_putlen1; // fix output 
	sei();
}

uint8_t i2cdev_can_putc(uint8_t len) {
	return I2CDEV_RX_BUFSIZE - i2cdev_rx_putlen1 >= len;
}

void i2cdev_putc(uint8_t c) {
	if (i2cdev_rx_putlen2) { // reset rx_buf if fixed
		cli();
		i2cdev_rx_putlen1 = 0;
		i2cdev_rx_putlen2 = 0;
		sei();
	}
	if (i2cdev_rx_putlen1 < I2CDEV_RX_BUFSIZE) {
		i2cdev_rx_buf[i2cdev_rx_putlen1++] = c;
	}
}

uint8_t i2cdev_can_getc(uint8_t len) {
	if (state == STATE_IDLE ||
		(state == STATE_RECV_DATA && bit_is_set(USISR, USIPF))) {
		i2cdev_tx_putlen2 = i2cdev_tx_putlen1;
	}
	return i2cdev_tx_putlen2 >= i2cdev_tx_getlen + len;
}

uint8_t i2cdev_getc() {
	while (!i2cdev_can_getc(1)) {
		;
	}
	return i2cdev_tx_buf[i2cdev_tx_getlen++];
}


static inline void set_idle() {
	bit_clear(I2CDEV_DDR, I2CDEV_SDA);
	USICR = (1<<USISIE)|(2<<USIWM0)|(2<<USICS0);
	USISR = (1<<USIOIF)|(1<<USIPF)|(1<<USIDC) | (0<<USICNT0);
}

static inline void send_ack() {
	USIDR = 0;
	bit_set(I2CDEV_DDR, I2CDEV_SDA);
	USISR = (1<<USIOIF)|(1<<USIPF)|(1<<USIDC) | (14<<USICNT0);
}

static inline void recv_reply() {
	bit_clear(I2CDEV_DDR, I2CDEV_SDA);
	USIDR = 0;
	USISR = (1<<USIOIF)|(1<<USIPF)|(1<<USIDC) | (14<<USICNT0);
}

static inline void send_data() {
	if (i2cdev_rx_putlen2 > i2cdev_rx_getlen) {
		USIDR = i2cdev_rx_buf[i2cdev_rx_getlen++];
	} else {
		USIDR = 0xff;
	}
	bit_set(I2CDEV_DDR, I2CDEV_SDA);
	USISR = (1<<USIOIF)|(1<<USIPF)|(1<<USIDC) | (0<<USICNT0);
}

static inline void recv_data() {
	bit_clear(I2CDEV_DDR, I2CDEV_SDA);
	USISR = (1<<USIOIF)|(1<<USIPF)|(1<<USIDC) | (0<<USICNT0);
}


void i2cdev_init(uint8_t myaddr) {
	i2cdev_tx_getlen = 0;
	i2cdev_tx_putlen1 = 0;
	i2cdev_tx_putlen2 = 0;
	i2cdev_rx_putlen1 = 0;
	i2cdev_rx_putlen2 = 0;

	USICR = 0;
	i2cdev_myaddr = myaddr<<1;
	bit_set(I2CDEV_PORT,I2CDEV_SCL);
	bit_set(I2CDEV_PORT,I2CDEV_SDA);
	bit_set(I2CDEV_DDR,I2CDEV_SCL);
	bit_clear(I2CDEV_DDR,I2CDEV_SDA);

	USISR = (1<<USISIF) | (1<<USIOIF) | (1<<USIPF) | (1<<USIDC);
	state = STATE_IDLE;
	/* enable I2C and USI_START interrupt enable */
	USICR = (1<<USISIE) | (2<<USIWM0) | (2<<USICS0);
}

SIGNAL(SIG_USI_START) {
	bit_clear(I2CDEV_DDR, I2CDEV_SDA);

	/* wait for start condition completed :
	 *
	 * SDA ~~~~~|_________________
         * SCL ~~~~~~~~~~~~~~~~~|______
	 *          @ now       @
	 *          |<--------->|
	 */
		 
	USICR &= ~((1<<USISIE) | (1<<USIOIE));
	sei();  // enable other interrupt (except USI_OVERFLOW,USI_START)
	while (bit_is_clear(USISR,USIPF) && bit_is_set(I2CDEV_PIN, I2CDEV_SCL))
		;
	cli();
	//usi_start_cond_time = get_time(); // for sync 
	state = STATE_RECV_ADDR;

	/* enable I2C (with SCL low) 
	 *    and USI_START/USI_OVERFLOW interrupt enable
	 */ 
	/* USICR change : set USIOIE (enable Overflow interrupt)
	 *              : USIWMO 2->3 (Low SCL until clear USIOIF)
	 */
	USICR = (1<<USISIE) | (1<<USIOIE) | (3<<USIWM0) | (2<<USICS0);
	USISR =  (1<<USISIF)|(1<<USIOIF)|(1<<USIPF)|(1<<USIDC) |(0<<USICNT0);
}

SIGNAL(SIG_USI_OVERFLOW) {
	switch (state) {
	case STATE_RECV_ADDR:
		if ((_USIBR == 0) || ((_USIBR & ~1) == i2cdev_myaddr)) {
			if (_USIBR & 1) { // read command
				i2cdev_rx_getlen = 0;
				state = STATE_SEND_DATA_START;
			} else { // write command
				if (i2cdev_tx_putlen2 > i2cdev_tx_getlen) {
					goto reinit; // no ACK
				}
				i2cdev_tx_putlen1 = 0;
				i2cdev_tx_putlen2 = 0;
				i2cdev_tx_getlen = 0;
				state = STATE_SEND_REPLY;
			}
			send_ack();
		} else {
			goto reinit;
		}
		break;
	case STATE_SEND_REPLY:
		state = STATE_RECV_DATA;
		recv_data();
		break;
	case STATE_RECV_DATA:
		if (i2cdev_tx_putlen1 < I2CDEV_TX_BUFSIZE) {
			i2cdev_tx_buf[i2cdev_tx_putlen1++] = _USIBR;
			state = STATE_SEND_REPLY;
			send_ack();
		}
		break;
	case STATE_SEND_DATA:
		recv_reply();
		state = STATE_RECV_REPLY;
		break;
	case STATE_RECV_REPLY:
		if ( _USIBR ) { // NACK : last data
			goto reinit;
		}
		// continue to STATE_SEND_DATA_START
	case STATE_SEND_DATA_START:
		send_data();
		state = STATE_SEND_DATA;
		break;
	}
	return;
reinit:
	state = STATE_IDLE;
	set_idle();
}
